#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2019 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.

""" Search small ad listings on OLX web sites in different countries """


import importlib


__all__ = [
    "OlxSearch"
]


class OlxSearch:
    """ Class factory for obtaining a country-specific class to scrape OLX
        market place web sites.
    """

    def __new__(
            cls,
            country,
            *args,
            **kwargs
    ):
        cls_ = "OlxSearch{:s}".format(country.replace(" ", ""))
        module_ = "olxsearch.lib.{:s}".format(cls_.lower())

        try:
            module = importlib.import_module(module_)
            cls = getattr(module, cls_)
        except ImportError as e:
            raise NotImplementedError(
                """ Could not find country-specific
                    OlxSearch class for {:s} """.format(country)
            ) from e

        instance = cls.__new__(cls, *args, **kwargs)
        instance.__init__(*args, **kwargs)

        return instance



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2019 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.

""" Search small ad listings on OLX web sites in several countries

    This is one of several models of OLX pages, and the one used in
        - Argentina
        - Colombia
        - Ecuador
        - Guatemala
        - India
        - Indonesia
        - Pakistan
        - Panama
        - Peru
        - San Salvador
        - South Africa
"""


import sys

import requests

from .olxsearchbase import OlxSearchBase

__all__ = [
    "OlxSearchBaseA"
]


class OlxSearchBaseA(OlxSearchBase):
    """ Base class for searching OLX listings in various countries,
        also see country-specific classes
    """

    @staticmethod
    def _filter_record(record):
        return {
            "id": record["id"],
            "title": record["title"],
            "description": record["description"],
            "created_at": record["created_at"],
            "created_at_first": record["created_at_first"],
            "republish_date": record["republish_date"],
            "images": [i["url"] for i in record["images"]],
            "price": (None, None) if record["price"] is None else (
                record["price"]["value"]["raw"],
                record["price"]["value"]["currency"]["iso_4217"]
            ),
            # "location_country":
            #     record["locations_resolved"]["COUNTRY_name"],
            # "location_a1":
            #     record["locations_resolved"]["ADMIN_LEVEL_1_name"],
            # "location_a3":
            #     record["locations_resolved"]["ADMIN_LEVEL_3_name"],
            # "location_sublocality":
            #     record["locations_resolved"]["SUBLOCALITY_LEVEL_1_name"],
            "lat": record["locations"][0]["lat"],
            "lon": record["locations"][0]["lon"]
        }

    def _download_listings(self):
        """ Download all listings for `search_term` """
        # don’t run the base class
        try:
            self.BASE_URL
        except AttributeError as e:
            raise NotImplementedError(
                "Use a country-specific OlxSearch class"
            ) from e

        page = 0
        while True:
            try:
                data = requests.get(
                    self.BASE_URL,
                    params={
                        "page": page,
                        "query": self.search_term
                    }
                ).json()
            except requests.RequestException as e:
                print(e, file=sys.stderr)
                break

            page += 1
            if not data["data"]:
                break

            for d in data["data"]:
                yield self._filter_record(d)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2019 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.

""" Search small ad listings on OLX web sites in several countries

    This is one of several models of OLX pages, and the one used in
        - Bosnia
"""


import json
import sys
import urllib.parse

import bs4
import dateparser.search
import fake_useragent
import geocoder
import requests

from .olxsearchbase import OlxSearchBase

__all__ = [
    "OlxSearchBaseE"
]


class OlxSearchBaseE(OlxSearchBase):
    """ Base class for searching OLX listings in various countries,
        also see country-specific classes
    """
    _useragent = fake_useragent.UserAgent(
        fallback="Mozilla/5.0 (Windows NT 10.0; Win64; x64) " +
        "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/70.0.3538.77 " +
        "Safari/537.36"
    ).random

    @classmethod
    def _filter_record(cls, record):
        filtered_record = {}
        try:
            details_href = \
                record.a["href"]

            with requests.get(
                    details_href,
                    headers={
                        "User-Agent": cls._useragent
                    }
            ) as response:
                details = bs4.BeautifulSoup(response.text, "html.parser")
                metadata = \
                    cls._find_js_metadata_in_details_page(details)

                filtered_record["title"] = \
                    record.find("p", class_="na").get_text()
                filtered_record["id"] = \
                    urllib.parse.urlparse(details_href).path.split("/")[2]

                try:
                    filtered_record["description"] = \
                        details\
                        .find("div", id="artikal_detaljniopis")\
                        .get_text()\
                        .replace("Detaljni opis", "")\
                        .strip()
                except AttributeError:
                    try:
                        filtered_record["description"] = \
                            details\
                            .find("div", id="detaljno")\
                            .get_text()\
                            .replace("Detaljni opis", "")\
                            .strip()
                    except AttributeError:
                        filtered_record["description"] = ""
                filtered_record["created_at"] = \
                    dateparser.search.search_dates(
                        record.find("div", class_="kada")["data-cijelidatum"],
                        languages=[cls.LANGUAGE]
                    )[0][1]
                filtered_record["created_at_first"] = None
                filtered_record["republish_date"] = None

                try:
                    filtered_record["images"] = [
                        image["href"]
                        for image
                        in metadata
                    ]
                except TypeError:
                    filtered_record["images"] = []

                try:
                    price = details\
                        .find("span", class_="markadesno")\
                        .parent\
                        .get_text()\
                        .replace(".", "")\
                        .replace("  ", " ")\
                        .strip()
                    filtered_record["price"] = price.split(" ")
                except AttributeError:
                    filtered_record["price"] = (None, None)

                try:
                    (lat, lon) = geocoder.osm(
                        details.find(class_="mobile-lokacija")["data-content"]
                    ).latlng
                except TypeError:
                    lat, lon = None, None

                filtered_record["lat"] = lat
                filtered_record["lon"] = lon

        except (
                KeyError,
                TypeError
        ):
            filtered_record = None

        return filtered_record

    @staticmethod
    def _find_js_metadata_in_details_page(details):
        metadata = {}
        for script in details.find_all("script"):

            if script.has_attr("src"):
                continue

            # expanding the filter of text types to
            # get to bs4.Comment and bs4.Script
            # Otherwise this does yields an empty string
            s_text = script.get_text(
                types=(
                    bs4.NavigableString,
                    bs4.CData,
                    bs4.Comment,
                    bs4.Script
                )
            )

            if "$.fancybox.open(" in s_text:
                for line in s_text.splitlines():
                    if "$.fancybox.open(" in line:
                        metadata = \
                            json.loads(
                                line[line.find("["):line.rfind("]") + 1]
                            )

        return metadata

    def _download_listings(self):
        """ Download all listings for `search_term` """
        # don’t run the base class
        try:
            self.BASE_URL
        except AttributeError as exception:
            raise NotImplementedError(
                "Use a country-specific OlxSearch class"
            ) from exception

        page = 1
        while True:
            try:
                params = {}
                if page > 1:
                    params["page"] = page

                data = requests.get(
                    self.BASE_URL.format(query=self.search_term),
                    headers={
                        "User-Agent": self._useragent
                    },
                    params=params
                ).text
            except requests.RequestException as exception:
                print(exception, file=sys.stderr)
                break

            data = bs4.BeautifulSoup(data, "html.parser")

            for record in data.find_all("div", class_="artikal"):
                record = self._filter_record(record)
                if record is None:
                    continue
                yield record

            # test whether there’s another page
            next_page_link = data.find("a", attrs={"rel": "next"})
            if next_page_link is None:
                break

            page += 1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2019 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.

""" Common base class for all OlxSearchX classes """


import pandas


__all__ = [
    "OlxSearchBase"
]


class OlxSearchBase:
    """ Base class for searching OLX listings in various countries,
        also see country-specific classes
    """

    def __init__(self, search_term=""):
        super().__init__()
        self.search_term = search_term

    @property
    def search_term(self):
        """ What to search for on OLX """
        return self._search_term

    @search_term.setter
    def search_term(self, value):
        self._search_term = value

    @property
    def listings(self):
        """ An Iterator over items posted for sale
            Raises StopIteration of no new posts available
        """
        try:
            self._listings
        except AttributeError:
            self._listings = self._download_listings()
        return self._listings

    @property
    def listings_as_dataframe(self):
        """
            A Pandas DataFrame containing all items posted for sale
        """
        return pandas.DataFrame(self.listings)

    @staticmethod
    def _filter_record(record):
        raise NotImplementedError("Use a country-specific OlxSearch class")

    def _download_listings(self):
        """ Download all listings for `search_term` """
        raise NotImplementedError("Use a country-specific OlxSearch class")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2019 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.

""" Search small ad listings on OLX web sites in several countries

    This is one of several models of OLX pages, and the one used in
        - Brazil
"""


import datetime
import json
import os
import sys
import time

import bs4
import fake_useragent
import geocoder
import requests

from .olxsearchbase import OlxSearchBase

__all__ = [
    "OlxSearchBaseD"
]


class OlxSearchBaseD(OlxSearchBase):
    """ Base class for searching OLX listings in various countries,
        also see country-specific classes
    """
    _useragent = fake_useragent.UserAgent(
        fallback="Mozilla/5.0 (Windows NT 10.0; Win64; x64) " +
        "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/70.0.3538.77 " +
        "Safari/537.36"
    ).random

    @classmethod
    def _filter_record(cls, record):
        filtered_record = {}
        try:
            details_href = record.find(
                "a",
                attrs={"data-lurker-detail": "list_id"}
            )["href"]

            with requests.get(
                    details_href,
                    headers={
                        "User-Agent": cls._useragent
                    }
            ) as response:
                details = bs4.BeautifulSoup(response.text, "html.parser")
                metadata = \
                    cls._find_js_metadata_in_details_page(details)

                offer = details.find("div", id="content")
                inner_container = \
                    offer\
                    .find_all("div", recursive=False)[1]\
                    .div\
                    .find_all("div", recursive=False)[1]\
                    .div

                filtered_record["id"] = metadata["listId"]
                filtered_record["title"] = \
                    details\
                    .find("title")\
                    .get_text()\
                    .replace(
                        " - {id:d} | OLX".format(id=filtered_record["id"]),
                        ""
                    )

                filtered_record["description"] = \
                    inner_container.find(
                        "span",
                        string="Descrição"
                    )\
                    .parent\
                    .parent\
                    .get_text()\
                    .replace("Descrição", "", 1)

                filtered_record["created_at"] = \
                    datetime.datetime.fromtimestamp(
                        metadata["page"]["detail"]["adDate"]
                    )
                filtered_record["created_at_first"] = None
                filtered_record["republish_date"] = None
                try:
                    filtered_record["images"] = [
                        img_item.div.img["src"]
                        for img_item
                        in offer.find_all(
                            "div",
                            attrs={"data-testid": "slides-wrapper"}
                        )
                    ]
                except TypeError:
                    filtered_record["images"] = []
                filtered_record["price"] = (
                    metadata["page"]["detail"]["price"],
                    "BRL"  # hard-coded because this base used for brazil only
                )

                municipality = \
                    inner_container\
                    .find("dt", string="Município")\
                    .parent\
                    .dd\
                    .get_text()
                try:
                    neighbourhood = \
                        inner_container\
                        .find("dt", string="Bairro")\
                        .parent\
                        .dd\
                        .get_text()
                except AttributeError:
                    neighbourhood = None
                postcode = \
                    inner_container\
                    .find("dt", string="CEP")\
                    .parent\
                    .dd\
                    .get_text()

                # EASTER EGG!
                # By default, `olxsearch` uses `geocoder.osm` to look up
                # geographic locations, deriving a coordinate pair from
                # neighbourhood and municipality names (in the case of Brazil)
                #
                # OLX ads also contain much preciser post code information.
                # `olxsearch` can also use cepaberto.com to retrieve geographic
                # coordinates for post codes. Register at cepaberto.com and
                # export your API token as an environment variable
                # CEPABERTO_TOKEN.
                #
                # While significantly more precise, this has one dramatic
                # downside, which is cepaberto.com’s rate limiting:
                # max. 1 location per second, max. 10,000 queries per day

                try:
                    assert "CEPABERTO_TOKEN" in os.environ
                    with requests.get(
                            "https://www.cepaberto.com/api/v3/cep",
                            headers={
                                "Authorization": "Token token={:s}".format(
                                    os.environ["CEPABERTO_TOKEN"]
                                )
                            },
                            params={"cep": postcode}
                    ) as response:
                        response = response.json()
                        lat = response["latitude"]
                        lon = response["longitude"]
                        time.sleep(1.0)  # crude way of respecting rate limit
                except (
                        AssertionError,
                        json.decoder.JSONDecodeError,
                        requests.RequestException
                ):
                    try:
                        (lat, lon) = geocoder.osm(
                            "{neighbourhood:s}, {municipality:s}".format(
                                neighbourhood=neighbourhood,
                                municipality=municipality
                            )
                        ).latlng
                    except TypeError:
                        try:
                            (lat, lon) = geocoder.osm(
                                municipality
                            ).latlng
                        except TypeError:
                            lat, lon = None, None

                filtered_record["lat"] = lat
                filtered_record["lon"] = lon

        except (
                KeyError,
                TypeError
        ):
            filtered_record = None

        return filtered_record

    @staticmethod
    def _find_js_metadata_in_details_page(details):
        metadata = {}
        for script in details.find_all("script"):

            if script.has_attr("src"):
                continue

            # expanding the filter of text types to
            # get to bs4.Comment and bs4.Script
            # Otherwise this does yields an empty string
            s_text = script.get_text(
                types=(
                    bs4.NavigableString,
                    bs4.CData,
                    bs4.Comment,
                    bs4.Script
                )
            )

            if "window.dataLayer = " in s_text:
                metadata = \
                    json.loads(
                        s_text[s_text.find("["):]
                    )[0]

        return metadata

    def _download_listings(self):
        """ Download all listings for `search_term` """
        # don’t run the base class
        try:
            self.BASE_URL
        except AttributeError as exception:
            raise NotImplementedError(
                "Use a country-specific OlxSearch class"
            ) from exception

        page = 1
        while True:
            try:
                params = {}
                if page > 1:
                    params["page"] = page

                data = requests.get(
                    self.BASE_URL.format(query=self.search_term),
                    headers={
                        "User-Agent": self._useragent
                    },
                    params=params
                ).text
            except requests.RequestException as exception:
                print(exception, file=sys.stderr)
                break

            data = bs4.BeautifulSoup(data, "html.parser")

            for record in data.find("ul", id="ad-list").find_all("li"):
                record = self._filter_record(record)
                if record is None:
                    continue
                yield record

            # test whether there’s another page
            next_page_link = \
                data.find("a", attrs={"data-lurker-detail": "next_page"})
            if next_page_link is None:
                break

            page += 1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright (C) 2019 Christoph Fink, University of Helsinki
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 3
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, see <http://www.gnu.org/licenses/>.

""" Search small ad listings on OLX Indonesia
    and save them in a local PostGIS/PostgreSQL database """


import argparse
import os
import os.path

import psycopg2
import psycopg2.extras
import psycopg2.sql
import requests

from ..lib.olxsearch import OlxSearch


def _save_listing_to_database(listing, connection_string, table_name):
    """ Save an OLX listing to a PostgreSQL database """
    with psycopg2.connect(connection_string) as connection:
        with connection.cursor(
                cursor_factory=psycopg2.extras.DictCursor
        ) as cursor:
            cursor.execute(
                psycopg2.sql.SQL(
                    """
                        INSERT INTO
                            {table_name}
                                (
                                    id,
                                    title,
                                    description,
                                    created_at,
                                    created_at_first,
                                    republish_date,
                                    price,
                                    price_currency,
                                    images,
                                    geom
                                )
                            VALUES (
                                %(id)s,
                                %(title)s,
                                %(description)s,
                                %(created_at)s,
                                %(created_at_first)s,
                                %(republish_date)s,
                                %(price)s,
                                %(price_currency)s,
                                %(images)s,
                                ST_SetSRID(
                                    ST_Point(%(lon)s, %(lat)s),
                                    4326
                                )
                            )
                        ON CONFLICT (id)
                            DO UPDATE
                                SET
                                    title=EXCLUDED.title,
                                    description=EXCLUDED.description,
                                    created_at=EXCLUDED.created_at,
                                    created_at_first=EXCLUDED.created_at_first,
                                    republish_date=EXCLUDED.republish_date,
                                    price=EXCLUDED.price,
                                    price_currency=EXCLUDED.price_currency,
                                    images=EXCLUDED.images,
                                    geom=EXCLUDED.geom;
                    """
                ).format(
                    table_name=psycopg2.sql.Identifier(table_name)
                ),
                {
                    "id": listing["id"],
                    "title": listing["title"],
                    "description": listing["description"],
                    "created_at": listing["created_at"],
                    "created_at_first": listing["created_at_first"],
                    "republish_date": listing["republish_date"],
                    "price": listing["price"][0],
                    "price_currency": listing["price"][1],
                    "images": listing["images"],
                    "lon": listing["lon"],
                    "lat": listing["lat"]
                }
            )


def _download_listing_images(listing, media_directory):
    """ Download all images attached to an OLX listing """
    destination_directory = os.path.join(
        media_directory,
        listing["id"]
    )
    os.makedirs(
        destination_directory,
        exist_ok=True
    )
    for url in listing["images"]:
        filename = os.path.join(
            destination_directory,
            "{:s}.jpg".format(
                url.split('/')[-2]
            )
        )
        _download_file(url, filename)


def _download_file(url, filename):
    with requests.get(
            url,
            stream=True
    ) as r:
        with open(filename, "wb") as f:
            for chunk in r.iter_content(chunk_size=None):
                f.write(chunk)


def _create_tables(connection_string, table_name):
    """ Create the tables to store OLX listings in a PostgreSQL database """
    with psycopg2.connect(connection_string) as connection:
        with connection.cursor(
                cursor_factory=psycopg2.extras.DictCursor
        ) as cursor:
            cursor.execute(
                psycopg2.sql.SQL(
                    """
                        CREATE EXTENSION IF NOT EXISTS
                            postgis;

                        CREATE TABLE IF NOT EXISTS
                            {table_name}
                        (
                            row BIGSERIAL PRIMARY KEY,
                            id BIGINT UNIQUE,
                            title TEXT,
                            description TEXT,
                            created_at TIMESTAMP,
                            created_at_first TIMESTAMP,
                            republish_date TIMESTAMP,
                            price NUMERIC,
                            price_currency CHAR(3),
                            images TEXT[],
                            geom GEOMETRY('POINT', 4326)
                        );

                        CREATE INDEX IF NOT EXISTS
                            {geom_index}
                        ON
                            {table_name}
                        USING
                            GIST(geom);

                    """
                ).format(
                    table_name=psycopg2.sql.Identifier(table_name),
                    geom_index=psycopg2.sql.Identifier(
                        "{:s}_geom_idx".format(table_name)
                    )
                )
            )
            connection.commit()


def main():
    """ Search small ad listings on OLX Indonesia
        and save them in a local PostGIS/PostgreSQL database """
    argparser = argparse.ArgumentParser()
    argparser.add_argument(
        "-c",
        "--country",
        help="""Query the OLX market place for COUNTRY.
                Specify more than one country using multiple switches
                (-c COUNTRY1 -c COUNTRY2) """,
        choices=(
            "Argentina",
            "Bulgaria",
            "Colombia",
            "Costa Rica",
            "Ecuador",
            "Guatemala",
            "India",
            "Indonesia",
            "Kazakhstan",
            "Pakistan",
            "Panama",
            "Peru",
            "Poland",
            "Portugal",
            "Romania",
            "San Salvador",
            "South Africa",
            "Ukraine",
            "Uzbekistan"
        ),
        action="append",
        required=True
    )
    argparser.add_argument(
        "search_terms",
        help="""Query the titles and descriptions of the listings
                on the OLX market place for this""",
        nargs="+"
    )
    argparser.add_argument(
        "-p",
        "--postgresql-connection-string",
        help="""Store the retrieved data in this PostgreSQL data base
                (default: "dbname=olx") """,
        default="dbname=olx"
    )
    argparser.add_argument(
        "-t",
        "--table",
        help="""Store the data in a PostgreSQL table with this name
                (default: same as the first search_term)""",
        default=None
    )
    argparser.add_argument(
        "-m",
        "--media-directory",
        help="""Download images to this directory""",
        default="./media"
    )
    args = argparser.parse_args()

    if args.table is None:
        args.table = args.search_terms[0]

    _create_tables(args.postgresql_connection_string, args.table)
    for search_term in args.search_terms:
        for country in args.country:
            for listing in OlxSearch(country, search_term).listings:
                _save_listing_to_database(
                    listing,
                    args.postgresql_connection_string,
                    args.table
                )
                _download_listing_images(
                    listing,
                    args.media_directory
                )


if __name__ == "__main__":
    main()
